from .api import connect
from .objdb import ObjDBMixin
from .dbobj import DbObj, stored
from .base import BaseDB

__all__ = [
    'connect',
    'stored',
    'ObjDBMixin',
    'DbObj',
    'stored',
    'BaseDB'
]



from .config import DB_MODULES
from .base import BaseDB
from .interface import KYDBInterface
from .union import UnionDB
from .cache import CacheDB
import importlib


_db_cache = {}


def connect(url: str) -> KYDBInterface:
    if "|" in url:
        dbs = [_connect(x) for x in url.split("|")]
        if len(dbs) != 2:
            raise ValueError("CacheDB expects exactly 2 databases")

        return CacheDB(*dbs)

    dbs = [_connect(x) for x in url.split(";")]
    if len(dbs) == 1:
        return dbs[0]
    return UnionDB(dbs)


def _connect(url: str) -> BaseDB:
    if url not in _db_cache:
        db_cls = _resolve_db_class(url)
        _db_cache[url] = db_cls(url)

    return _db_cache[url]


def _resolve_db_class(url: str):
    db_type = url.split(":", 1)[0]
    assert db_type in DB_MODULES, "{} is not one of the valid db types: {}".format(
        db_type, list(iter(DB_MODULES.keys()))
    )

    class_name = DB_MODULES[db_type]
    module_path = __name__.rsplit(".", 1)[0] + ".impl." + db_type
    m = importlib.import_module(module_path)
    return getattr(m, class_name)



from kydb.base import BaseDB
import pathlib
import os
import os.path


class FileDB(BaseDB):
    """
    Example::

    db = kydb('files://tmp/foo/bar') # must be absolute path

    would read and write files under /tmp/foo/bar
    """

    def __init__(self, url: str):
        """
        :param url: str: the URL starting with file://
        """
        super().__init__(url)

    def get_raw(self, key: str):
        try:
            return open(self._get_fs_path(key), 'rb').read()
        except FileNotFoundError:
            raise KeyError(key)

    def mkdir_raw(self, folder: str):
        folder = self._get_fs_path(folder)
        pathlib.Path(folder).mkdir(parents=True, exist_ok=True)

    def set_raw(self, key: str, value):
        """
        save value to file system

        :param key: str: the path to the file including base_path,
                         but excluding the db_name. e.g.
                         for kydb.connect(files://tmp/foo/bar)
                         key would be /foo/bar
        :param value: The raw, pickled data.
        """
        fullpath = self._get_fs_path(key)
        folder = fullpath.rsplit('/', 1)[0]
        pathlib.Path(folder).mkdir(parents=True, exist_ok=True)
        with open(fullpath, 'wb') as f:
            f.write(value)

    def delete_raw(self, key: str):
        """
        Delete a the file from filesystem
        To be implemented by derived class

        :param key: str: the path to the file including base_path,
                         but excluding the db_name. e.g.
                         for kydb.connect(files://tmp/foo/bar)
                         key would be /foo/bar

        """
        os.remove(self._get_fs_path(key))

    def _get_fs_path(self, key: str):
        return '/' + self.db_name + key

    def is_dir_raw(self, folder: str) -> bool:
        """ Is this a directory?

        :param folder: Returns True if is directory
        """
        path = self._get_fs_path(folder)
        return os.path.isdir(path)

    def list_dir_raw(self, folder: str, include_dir: bool, page_size: int):
        folder = self._get_fs_path(folder)
        try:
            for filename in os.listdir(folder):
                path = self._ensure_slashes(folder) + filename
                if os.path.isdir(path):
                    if include_dir:
                        yield filename + '/'
                else:
                    yield filename
        except FileNotFoundError:
            raise KeyError(folder)

    def rmdir_raw(self, folder: str):
        try:
            os.rmdir(self._get_fs_path(folder))
        except OSError:
            raise KeyError('Cannot remove folder: ' + folder)

    def exists_raw(self, key) -> bool:
        path = self._get_fs_path(key)
        return os.path.exists(path) and not os.path.isdir(path)



from kydb.base import BaseDB
from botocore.exceptions import ClientError, ParamValidationError
import io
import boto3
from kydb.folder_meta import FolderMetaMixin


class S3DB(FolderMetaMixin, BaseDB):

    def __init__(self, url: str):
        super().__init__(url)
        self.s3 = boto3.client('s3')

    def get_raw(self, key: str):
        try:
            buf = io.BytesIO()
            self.s3.download_fileobj(self.db_name, key[1:], buf)
            return buf.getvalue()
        except ClientError:
            raise KeyError(key)
        except ParamValidationError:
            raise KeyError(key)

    def folder_meta_set_raw(self, key: str, value):
        buf = io.BytesIO(value)
        self.s3.upload_fileobj(buf, self.db_name, key[1:])

    def delete_raw(self, key: str):
        self.s3.delete_object(
            Bucket=self.db_name,
            Key=key[1:]
        )

    def list_dir_meta_folder(self, folder: str, page_size: int):
        """ List the folder

        :param folder: The folder to lsit
        :parm include_dir: include subfolders
        :parm page_size: The number of items to fetch at a time from DB
                         The result would be identical, only controls
                         performance

        Note Folders always ends with ``/``
        Objects does not
        """
        folder = self._ensure_slashes(folder)[1:]

        res = {}

        while True:
            kwargs = {
                'Bucket': self.db_name,
                'Delimiter': '/',
                'MaxKeys': page_size,
                'Prefix': folder
            }

            token = res.get('NextContinuationToken')

            if token:
                kwargs['ContinuationToken'] = token

            res = self.s3.list_objects_v2(**kwargs)

            for item in res.get('Contents', []):
                path = item['Key']
                yield path.rsplit('/', 1)[1] if '/' in path else path

            if not res['IsTruncated']:
                break




from .base import BaseDB
from .interface import KYDBInterface
from typing import Tuple
from contextlib import ExitStack


def front_db_func(self, func_name, *args, **kwargs):
    return getattr(self.dbs[0], func_name)(*args, **kwargs)


def first_success_db_func(self, func_name, *args, **kwargs):
    first_error = None
    for db in self.dbs:
        try:
            return getattr(db, func_name)(*args, **kwargs)
        except KeyError as err:
            if not first_error:
                first_error = err
            continue

    raise first_error


def any_db_func(self, func_name, *args, **kwargs):
    return any(getattr(db, func_name)(*args, **kwargs) for db in self.dbs)


def all_db_func(self, func_name, *args, **kwargs):
    for db in self.dbs:
        getattr(db, func_name)(*args, **kwargs)


def create_func(func_prototype, func_name):
    # Using partial loses the self when constructing
    # class using type. So use this function
    def f(self, *args, **kwargs):
        return func_prototype(self, func_name, *args, **kwargs)

    return f


UNION_DB_BASE_FUNCS = [
    ('__getitem__', first_success_db_func),
    ('__setitem__', front_db_func),
    ('delete', front_db_func),
    ('rmdir', front_db_func),
    ('rm_tree', front_db_func),
    ('new', front_db_func),
    ('exists', any_db_func),
    ('refresh', all_db_func),
    ('read', first_success_db_func),
    ('mkdir', front_db_func),
    ('is_dir', any_db_func),
    ('upload_objdb_config', front_db_func)
]

UnionDBBase = type(
    'UnionDBBase',
    (KYDBInterface,),
    {k: create_func(v, k) for k, v in UNION_DB_BASE_FUNCS}
)


class UnionDB(UnionDBBase):
    """UnionDB


The URL used on *connect* can be a semi-colon separated string.

This would create a Union Database.

Connecting::

    db = kydb.connect('memory://unittest;s3://my-unittest-fixture')

OR::

    db = kydb.connect('redis://hotfixes.epythoncloud.io;'
                      'dynamodb://my-prod-src-db')

Reading and writing::

    db1, db2 = db.dbs
    db1['/foo'] = 1
    db2['/bar'] = 2

    (db['/foo'], db['/bar']) # return (1, 2)

    # Although db2 has /foo, it is db1's /foo that the union returns
    db2['/foo'] = 3
    db['/foo'] # return 1

    # writing always happens on the front db
    db['/foo'] = 4
    db1['/foo'] # returns 4
    db2['/foo'] # returns 3
    """

    def __init__(self, dbs: Tuple[BaseDB]):
        self.dbs = dbs

    def cache_context(self) -> 'KYDBInterface':
        with ExitStack() as stack:
            for db in self.dbs:
                stack.enter_context(db.cache_context())

        return stack

    def list_dir(self, folder: str, include_dir=True, page_size=200):
        res = set()
        for db in self.dbs:
            try:
                res.update(db.list_dir(folder, include_dir, page_size))
            except KeyError:
                pass

        for key in res:
            yield key

    def ls(self, folder: str, include_dir=True):
        return list(self.list_dir(folder, include_dir))

    def __repr__(self):
        """
        The representation of the db.

        i.e. <UnionDB redis://my-redis-host/source;
              kydb.S3 s3://my-s3-prod-source>
        """
        return f'<{type(self).__name__} ' + \
            ';'.join(db.url for db in self.dbs) + '>'



from abc import ABC


class KYDBInterface(ABC):
    """The interface that all KYDB adheres to

    This class and all derived classes are never instantiated directly.

Instead use the ``connect``. i.e.::

    import kydb

    db = kydb.connect('dynamodb://my-table')

    """

    def __getitem__(self, key: str):
        """Get data from the DB based on key

        To be implemented by derived class

        :param key: str:  The key to get.

example::

    db[key] # returns the object with key

        """
        raise NotImplementedError()

    def __setitem__(self, key: str, value):
        """
        Set data from the DB based on key

        :param key: str:  The key to set.
        :param value:  The python object

example::

    db[key] = value # sets key to value

Note: key cannot have any component that has
a dot (.) prefix.

i.e. the below are illegal and would raise KeyError

::

    db['.foo'] = 123 # raises KeyError
    db['/.foo'] = 123 # raises KeyError
    db['/my/folder/.foo'] 123 # raises KeyError
    db['/my-folder/.another-folder/foo'] = 123 # raises KeyError

        """
        raise NotImplementedError()

    def set(self, key: str, value, system_obj=False):
        """Set data from the DB based on key

        :param key: str:  The key to set.
        :param value:  The python object
        :param system_obj: bool: True if a system object

        Same as __setitem__ except it can write system objects
        i.e. object with a (.) dot prefix.

        Note: only use this if you know what you're doing
        """
        raise NotImplementedError()

    def list_dir(self, folder: str, include_dir=True, page_size=200):
        """ List the folder

        :param folder: The folder to lsit
        :parm include_dir: include subfolders
        :parm page_size: The number of items to fetch at a time from DB
                         The result would be identical, only controls
                         performance

        Note Folders always ends with ``/``
        Objects does not
        """
        raise NotImplementedError()

    def ls(self, folder: str, include_dir=True):
        """ Similar to list_dir, but returns a list (not generator)

        :param folder: The folder to lsit
        :parm include_dir: include subfolders
        :parm page_size: The number of items to fetch at a time from DB
                         The result would be identical, only controls
                         performance

        Note Folders always ends with ``/``
        mbjects does not
        """
        raise NotImplementedError()

    def delete(self, key: str):
        """
        Delete a key from the db.
        To be implemented by derived class

        :param key: str:  The key to delete.

example::

    db.delete(key) # Deletes data with key
        """
        raise NotImplementedError()

    def rmdir(self, key: str):
        """
        Delete folder based on key

        :param key: str:  The key to folder to delete

example::

    db.rmdir(folder) # Deletes folder with key
        """
        raise NotImplementedError()

    def rm_tree(self, key: str):
        """ recursively delete folder

.. warning::

    Be careful when using this.
    For example ``rm_tree('/')`` would wipe out the entire database!

        """
        raise NotImplementedError()

    def new(self, class_name: str, key: str, **kwargs):
        """
        Create a new object on the DB.
        The object is not persisted until obj.write() is called.

        :param class_name: str: name of the class.
                                This name must be in the config registry
        :param key: str: The key to persist on the DB
        :param kwargs: the stored attributes to set on the obj
        :returns: an obj of type defined by class_name

example::

    obj = db.new('MyClass', key, foo=3)
        """
        raise NotImplementedError()

    def exists(self, key) -> bool:
        """
        Check if a key exists in the DB

        :param key: the key
        :returns: True if key exists, False otherwise.

        Example:

::

    db['/my/key'] = 123
    db.exists('/my/key') # returns True

        """
        raise NotImplementedError()

    def refresh(self, key=None):
        """
        Flush the cache

        :param key: Optionally choose which key to flush (Default value = None)

example::

    obj = db.new('MyClass', key)
    obj.write()
    db[key] # read from cache
    db.refresh() # Or db.refresh(key)
    db[key] # read from DB

        """
        raise NotImplementedError()

    def read(self, key: str, reload=False):
        """
        Read object from DB given the key.
        If key has been read before, this call would simply
        return the cached value. Use reload to force reloading
        from DB.

        :param key: str: key to DB
        :param reload:  Optionally force reloading of the object
                        from db (Default value = False)
        :returns: The object from DB

example::

    obj = db.new('MyClass', key)
    obj.write()
    db.read(key) # read from cache
    db.read(key, reload=True) # Force loading from DB

        """
        raise NotImplementedError()

    def mkdir(self, folder: str):
        """ Make a directory (recursively if required)

        :param folder: The folder path.

example::

    db.mkdir('/foo/bar')
    db.ls('/foo') # returns ['bar/']

        """

    def is_dir(self, folder: str) -> bool:
        """ Is this a directory?

        :param folder: Returns True if is directory

example::

    db.mkdir('/foo/bar')
    db.is_dir('/foo/bar') # returns True

        """
        raise NotImplementedError()

    def cache_context(self) -> 'KYDBInterface':
        """ returns the cache context

        See :ref:`Cache Context`
        """
        raise NotImplementedError()

    def __repr__(self):
        """ The representation of the db.

        kydb.connect('s3://my-db')
        # displays <S3DB s3://my-db>

        kydb.connect('redis://my-cache;dynamodb://my-db')
        # displays <UnionDB redis://my-cache;dynamodb://my-db>
        """
        raise NotImplementedError()

    def upload_objdb_config(self, config):
        """Upload ObjDB config to KYDB

        :param config: The config dict

           This should only need to be done when new classes are registered or
           existing ones changes path.

::

    db = kydb.connect('memory://decorated_py_obj')

    db.upload_objdb_config({
        'Greeter': {
            'module_path': 'path.to.module',
            'class_name': 'Greeter'
        }
    })
        """
        raise NotImplementedError()

    def clear_cache(self):
        """Clear the cache

        This is useful when you want to clear the cache from memory

        Note: This is different to CacheDB where the cache is a database
        """
        raise NotImplementedError()



from contextlib import contextmanager
import copy


@contextmanager
def cache_context(db):
    """The Cache Context

    See :ref:`Cache Context`

    """
    orig_cache = db._cache
    db._cache = copy.copy(orig_cache)

    try:
        yield db
    finally:
        db._cache = orig_cache



from kydb.base import BaseDB
from kydb.folder_meta import FolderMetaMixin
import re


class MemoryDB(FolderMetaMixin, BaseDB):
    __cache = {}

    def __init__(self, url: str):
        super().__init__(url)
        self.__cache[self.db_name] = {}

    def get_raw(self, key):
        if self.base_path != '/' and \
                key == self._folder_meta_path(self.base_path, ''):
            raise KeyError(key)

        return self.__cache[self.db_name][key]

    def folder_meta_set_raw(self, key: str, value):
        self.__cache[self.db_name][key] = value

    def delete_raw(self, key: str):
        del self.__cache[self.db_name][key]

    def get_cache(self):
        return self.__cache[self.db_name]

    @staticmethod
    def _list_dir_regex(folder: str):
        return re.compile(f'^{folder}([^/]+)$')

    def list_dir_meta_folder(self, folder: str, page_size: int):
        folder = self._ensure_slashes(folder)
        pattern = self._list_dir_regex(folder)
        for path in self.__cache[self.db_name].keys():
            match = pattern.search(path)
            if match:
                yield match.groups(0)[0]



from kydb.base import BaseDB
from kydb.folder_meta import FolderMetaMixin
from redis.exceptions import ResponseError
import redis
import boto3
import os
import base64


class RedisDB(FolderMetaMixin, BaseDB):

    def __init__(self, url: str):
        super().__init__(url)

        self.connection = redis.Redis(
            **self._get_connection_kwargs(self.db_name))

    def _get_connection_kwargs(self, db_name: str):
        if self._config:
            password = self._get_password()
            host = self._config['host']
            port = self._config['port']
            return {
                'host': host,
                'port': port,
                'password': password,
            }

        return self._get_connection_from_dbname(db_name)

    def _get_password(self):
        pwd_cfg = self._config['password']
        encrypt_method = pwd_cfg['encryption-method']
        if encrypt_method == 'kms':
            return self._get_secret_from_kms(pwd_cfg['env_var'], pwd_cfg['encryption-key'])

        if encrypt_method == 'plain':
            return os.environ[pwd_cfg['env_var']]

        raise ValueError(f'Unknown encryption method: {encrypt_method}')

    @staticmethod
    def _get_secret_from_kms(name, kms_key_id: str):
        kms = boto3.client('kms')
        encrypted = os.environ[name]
        res = kms.decrypt(
            KeyId=kms_key_id,
            CiphertextBlob=base64.b64decode(encrypted))

        return res['Plaintext'].decode()

    @staticmethod
    def encrypt_secret(secret: str, kms_key_id: str):
        kms = boto3.client('kms')
        res = kms.encrypt(
            KeyId=kms_key_id,
            Plaintext=secret)
        return base64.b64encode(res['CiphertextBlob'])

    @staticmethod
    def _get_connection_from_dbname(db_name: str):
        if ':' in db_name:
            host, port = db_name.split(':')
            kwargs = {
                'host': host,
                'port': int(port, 10)
            }
        else:
            kwargs = {
                'host': db_name
            }

        return kwargs

    def get_raw(self, key: str):
        try:
            res = self.connection.get(key)
        except ResponseError:
            raise KeyError(f'{key} is not a valid key')

        if not res:
            raise KeyError(key)

        return res

    def folder_meta_set_raw(self, key: str, value):
        folder, obj = key.rsplit('/', 1)
        self.connection.hset(folder, obj, '.')
        self.connection.set(key, value)

    def delete_raw(self, key: str):
        self.connection.delete(key)
        folder, obj = key.rsplit('/', 1)
        self.connection.hdel(folder, obj)

    def list_dir_meta_folder(self, folder: str, page_size: int):
        folder = self._ensure_slashes(folder)[:-1]
        try:
            for key in self.connection.hgetall(folder).keys():
                yield key.decode()
        except ResponseError:
            raise KeyError(f'{folder} is not a valid folder')



from kydb.base import BaseDB
import requests


class HttpDB(BaseDB):
    """
    HTTP implementation of KYDB.

    Currently supports read only

    However if for example the HTTP query is requesting from a static web host
    on an S3 bucket. Then the s3 KYDB can write to it and then reading can
    be done with this same class.
    """

    def __init__(self, url: str):
        super().__init__(url)

    def get_raw(self, key: str):
        r = requests.get('{}://{}{}'.format(self.db_type, self.db_name,  key))
        if not r.ok:
            raise KeyError(key)

        return bytes.fromhex(r.text)



from kydb.base import BaseDB
from boto3.dynamodb.conditions import Key
from kydb.folder_meta import FolderMetaMixin
import boto3


class DynamoDB(FolderMetaMixin, BaseDB):

    def __init__(self, url: str):
        super().__init__(url)
        dynamodb = boto3.resource('dynamodb')
        self.table = dynamodb.Table(self.db_name)

    def get_raw(self, key):
        items = self.table.query(
            KeyConditionExpression=Key('path').eq(key))['Items']

        if not items:
            raise KeyError(key)

        return items[0]['contents'].value

    def folder_meta_set_raw(self, key: str, value):
        folder = key.rsplit('/', 1)[0] + '/'
        self.table.put_item(Item={
            'path': key,
            'folder': folder,
            'contents': value
        })

    def delete_raw(self, key: str):
        self.table.delete_item(Key={
            'path': key,
        })

    def list_dir_meta_folder(self, folder: str, page_size: int):
        folder = self._ensure_slashes(folder)

        done = False
        start_key = None
        while not done:
            kwargs = {
                'IndexName': "folder-index",
                'KeyConditionExpression': Key('folder').eq(folder),
                'Limit': page_size
            }

            if start_key:
                kwargs['ExclusiveStartKey'] = start_key

            res = self.table.query(**kwargs)

            for item in res.get('Items', []):
                yield item['path'].rsplit('/', 1)[1]

            start_key = res.get('LastEvaluatedKey', None)
            done = start_key is None



import pickle
from typing import Tuple
import os
from .objdb import ObjDBMixin
from .cache_context import cache_context
from .interface import KYDBInterface
from typing import Optional
import yaml


class BaseDB(ObjDBMixin, KYDBInterface):
    """ Base class for KYDBInterface """

    def __init__(self, url: str):
        self.db_type = url.split(':', 1)[0]
        self.db_name, self.base_path = self._get_name_and_basepath(url)
        self._config = self._get_config()
        self.url = url
        self._cache = {}

    def _get_config(self) -> Optional[dict]:
        config_path = os.environ.get('KYDB_CONFIG_PATH')
        if config_path:
            with open(config_path, 'r') as f:
                config = yaml.safe_load(f)

            return config['dbs'].get(self.db_name)

    @staticmethod
    def _get_name_and_basepath(url: str) -> Tuple[str, str]:
        """
        get the db_name and base_path

        :param url: str:
        :returns: a tuple (db_name, base_path)

        """
        base_path = '/'
        parts = url.split('/', 3)
        num_parts = len(parts)
        assert num_parts in (3, 4)

        db_name = parts[2]
        if len(parts) == 4:
            base_path += parts[-1]

        if base_path[-1] != '/':
            base_path += '/'

        return db_name, base_path

    def _get_full_path(self, key: str) -> str:
        """
        get the fullpath from key including base_path.

        :param key: str:
        :returns: The fullpath from key including base_path.

        """
        if key.startswith('/'):
            key = key[1:]

        res = self.base_path + key

        if not res.startswith('/'):
            res = '/' + res

        return res

    def _serialise(self, obj):
        """
        Serialises the object.

        :param obj: object to serialise
        :returns: The obj pickled

        """
        return pickle.dumps(obj)

    def _deserialise(self, data):
        """ Deserialise the data

        :param obj:
        :returns: Unpickle data
        """
        return pickle.loads(data)

    def exists(self, key) -> bool:
        """ Implements exists in KYDBInterface """
        return self.exists_raw(self._get_full_path(key))

    def exists_raw(self, key: str) -> bool:
        """ Same as exist but with base_path prepended """
        try:
            self.get_raw(key)
            return True
        except KeyError:
            return False

    def __getitem__(self, key: str):
        """ Implements __getitem__ in KYDBInterface """
        return self.read(key)

    def refresh(self, key=None):
        """ Implements refresh in KYDBInterface """
        if key:
            del self._cache[key]
        else:
            self._cache = {}

    def clear_cache(self):
        """Clear the cache

        This is useful when you want to clear the cache from memory

        Note: This is different to CacheDB where the cache is a database
        """
        self._cache = {}

    def read(self, key: str, reload=False):
        """ Implements read in KYDBInterface """
        path = self._get_full_path(key)
        res = None if reload else self._cache.get(path)
        if not res:
            res = self._deserialise(self.get_raw(path))
            if self.is_data_dbobj(res):
                res = self.read_dbobj(res)

        self._cache[key] = res
        return res

    def mkdir(self, folder: str):
        """ Implements read in KYDBInterface """
        if not folder or folder == '/':
            raise ValueError('Cannot make folder: ' + folder)

        self.mkdir_raw(self._get_full_path(folder))

    def mkdir_raw(self, folder: str):
        """ same as mkdir but with base_path prepended """
        raise NotImplementedError()

    def is_dir(self, folder: str) -> bool:
        return self.is_dir_raw(self._get_full_path(folder))

    def is_dir_raw(self, folder: str) -> bool:
        """ Same as is_dir, but prepended with base_path """
        raise NotImplementedError()

    def __setitem__(self, key: str, value):
        self.set(key, value)

    def set(self, key: str, value, system_obj=False):
        if not system_obj and \
                any(x for x in key.rsplit('/') if x.startswith('.')):
            raise KeyError('Cannot have dot (.) prefix in path, '
                           'got :' + key)

        path = self._get_full_path(key)
        self._cache[path] = value

        if self.is_dbobj(value):
            self.write_dbobj(value)
        else:
            self.set_raw(path, self._serialise(value))

    def get_raw(self, key: str):
        """
        Get data from the DB based on key.

        This is to be implemented by derived class.

        :param key: str:  The key to get, including base_path.
        :returns: str: The raw, pickled data.
        """
        raise NotImplementedError()

    def set_raw(self, key: str, value):
        """
        Set data from the DB based on key.

        This is to be implemented by derived class.

        :param key: str:  The key to set, including base_path.
        :param value: The raw, pickled data.
        """
        raise NotImplementedError()

    def delete_raw(self, key: str):
        """
        Delete data from the DB based on key

        This is to be implemented by derived class.

        :param key: str:  The key to delete

        """
        raise NotImplementedError()

    def delete(self, key: str):
        if not self.exists(key):
            raise KeyError('Cannot delete non-existence: ' + key)

        if key in self._cache:
            del self._cache[key]

        self.delete_raw(self._get_full_path(key))

    def rmdir(self, key: str):
        if key in ['.', '/', '']:
            raise KeyError('Directory does not exist: ' + key)

        path = self._get_full_path(key)

        if not self.exists_raw(path) and not self.is_dir_raw(path):
            raise KeyError('Directory does not exist: ' + path)

        try:
            next(iter(self.list_dir(key, page_size=1)))
            raise KeyError(f'Directory {key} is not empty')
        except StopIteration:
            self.rmdir_raw(path)

    def rmdir_raw(self, key: str):
        """ same as rmdir but with base_path prepended """
        raise NotImplementedError()

    def list_dir(self, folder: str, include_dir=True, page_size=200):
        return self.list_dir_raw(
            self._get_full_path(folder), include_dir, page_size)

    def list_dir_raw(self, folder: str, include_dir: bool, page_size: int):
        """ Same as list_dir but with base_path prepended to folder """
        raise NotImplementedError()

    def ls(self, folder: str, include_dir=True):
        return list(self.list_dir(folder, include_dir))

    def rm_tree(self, key: str):
        if not self.is_dir(key):
            raise KeyError('{} is not a directory'.format(key))

        folder = self._ensure_slashes(key)
        objs = list(self.list_dir(key))
        for obj in objs:
            path = folder + obj
            if obj.endswith('/'):
                self.rm_tree(path)
            else:
                self.delete(path)

        self.rmdir(key)

    def new(self, class_name: str, key: str, **kwargs):
        return self.db_obj_new(class_name, key, kwargs)

    def cache_context(self) -> KYDBInterface:
        return cache_context(self)

    def __repr__(self):
        """
        The representation of the db.

        i.e. <kydb.RedisDB redis://my-redis-host/source>
        """
        return f'<{type(self).__name__} {self.url}>'

    @staticmethod
    def _ensure_slashes(s: str):
        """Ensures s starts with / and ends with /

        :param s: The string to pass in

        Add slash in front or behind if needed
        """
        if not s.endswith('/'):
            s += '/'

        if not s.startswith('/'):
            s = '/' + s

        return s



class FolderMetaMixin:
    """ Used for providing list dir on a folder
        for DB implementations that does not
        have such mechanics
    """

    def mkdir_raw(self, folder: str):
        folders = self._ensure_slashes(folder)[1:-1].split('/')
        curr_folder = '/'
        for folder in folders:
            meta_path = self._folder_meta_path(curr_folder, folder)
            if not self.exists(meta_path):
                self.set_raw(meta_path, self._serialise(True))

            curr_folder += folder + '/'

    @classmethod
    def _folder_meta_path(cls, folder: str, subfolder: str = ''):
        folder = cls._ensure_slashes(folder)

        if not subfolder:
            folder = folder[:-1]
            if '/' not in folder:
                raise KeyError('Bad folder: ' + folder)

            folder, subfolder = folder.rsplit('/', 1)
            folder += '/'

        return folder + '.folder-' + subfolder

    @staticmethod
    def _is_folder_meta(objname: str):
        return objname.startswith('.folder-')

    def list_dir_raw(self, folder: str, include_dir: bool, page_size: int):
        for objname in self.list_dir_meta_folder(folder, page_size):
            if self._is_folder_meta(objname):
                if include_dir:
                    # Cut out the folder meta prefix
                    # and add /
                    yield objname[8:] + '/'
            else:
                yield objname

    def list_dir_meta_folder(self, folder: str, page_size: int):
        raise NotImplementedError()

    def is_dir_raw(self, folder: str) -> bool:
        return self.exists_raw(self._folder_meta_path(folder))

    def rmdir_raw(self, folder: str):
        return self.delete_raw(self._folder_meta_path(folder))

    def set_raw(self, key: str, value):
        key = self._ensure_slashes(key)[:-1]
        folder = key.rsplit('/', 1)[0]

        if folder:
            self.mkdir_raw(folder)
        self.folder_meta_set_raw(key, value)

    def folder_meta_set_raw(self, key: str, value):
        raise NotImplementedError()



from .base import BaseDB
from .interface import KYDBInterface
from contextlib import ExitStack
from .objdb import ObjDBMixin, DBOBJ_CONFIG_PATH
from .dbobj import DbObj
import pickle


class CacheDB(KYDBInterface, ObjDBMixin):
    """CacheDB
    """

    def __init__(self, cache_db: BaseDB, persist_db: BaseDB):
        self.cache_db = cache_db
        self.persist_db = persist_db

    def cache_context(self) -> 'KYDBInterface':
        """This is related to in memory cache

        Not to be confused with the cache_db which is
        remote. For example Redis
        """
        with ExitStack() as stack:
            stack.enter_context(self.cache_db.cache_context())
            stack.enter_context(self.persist_db.cache_context())

        return stack

    def list_dir(self, folder: str, include_dir=True, page_size=200):
        """List directory always looks at the persist_db"""
        yield from self.persist_db.list_dir(folder, include_dir, page_size)

    def ls(self, folder: str, include_dir=True):
        return list(self.list_dir(folder, include_dir))

    def __repr__(self):
        """
        The representation of the db.

        i.e. <CacheDB redis://my-redis-host/source;s3://my-s3-prod-source>
        """
        return f'<{type(self).__name__} {self.cache_db.url};{self.persist_db.url}>'

    def __getitem__(self, key: str):
        """Get item from DB

        Same as ``read(key)``"""
        return self.read(key)

    def __setitem__(self, key: str, value):
        """Write the item in both cache_db and persist_db

        Warning: If cache_db writes successfully and persist_db fails
        the two dbs will be out of sync
        """
        self.cache_db[key] = value
        self.persist_db[key] = value

    def delete(self, key: str):
        """Delete the item in both cache_db and persist_db

        Warning: If cache_db deletes successfully and persist_db fails
        the two dbs will be out of sync
        """
        self.cache_db.delete(key)
        self.persist_db.delete(key)

    def rmdir(self, key: str):
        """Remove the directory in both cache_db and persist_db

        Warning: If cache_db deletes successfully and persist_db fails
        the two dbs will be out of sync
        """
        self.cache_db.rmdir(key)
        self.persist_db.rmdir(key)

    def rm_tree(self, key: str):
        """Remove the directory recursively in both cache_db and persist_db

        Warning: If cache_db deletes successfully and persist_db fails
        the two dbs will be out of sync
        """
        self.cache_db.rm_tree(key)
        self.persist_db.rm_tree(key)

    def new(self, class_name: str, key: str, **kwargs):
        return self.db_obj_new(class_name, key, kwargs)

    def exists(self, key) -> bool:
        """Check if a key exists in either cache_db or persist_db"""
        return self.cache_db.exists(key) or self.persist_db.exists(key)

    def refresh(self, key=None):
        """Refresh both cache_db and persist_db"""
        self.cache_db.refresh(key)
        self.persist_db.refresh(key)

    def read(self, key: str, reload=False):
        """Get Item from CacheDB

        Try to get the item from the cache_db first
        If it does not exist, get it from the persist_db
        and then write it to the cache_db
        """

        def _ensure_db(obj):
            if isinstance(obj, DbObj):
                obj.db = self

            return obj

        if self.cache_db.exists(key):
            raw_data = self.cache_db.get_raw(key)

            data = pickle.loads(raw_data)
            if self.is_data_dbobj(data):
                m = data['meta']
                return self.db_obj_new(m['class_name'], m['key'], data['data'])

            return _ensure_db(self.cache_db[key])

        item = _ensure_db(self.persist_db[key])
        self.cache_db[key] = item

        return item

    def mkdir(self, folder: str):
        """Apply mkdir to both cache_db and persist_db"""
        self.cache_db.mkdir(folder)
        self.persist_db.mkdir(folder)

    def is_dir(self, folder: str) -> bool:
        """Check is_dir in persist db"""
        return self.persist_db.is_dir(folder)

    def upload_objdb_config(self, objdb_config: dict):
        """Upload objdb_config to both cache_db and persist_db"""
        self.cache_db.upload_objdb_config(objdb_config)
        self.persist_db.upload_objdb_config(objdb_config)

    def _get_dbobj_config(self, class_name: str):
        if not self.cache_db.exists(DBOBJ_CONFIG_PATH):
            self.cache_db.upload_objdb_config(
                self.persist_db[DBOBJ_CONFIG_PATH])

        return super()._get_dbobj_config(class_name)

    def clear_cache(self):
        """Clear the cache

        This is useful when you want to clear the cache from memory

        Note: This is different to CacheDB where the cache is a database
        """
        self.cache_db.clear_cache()
        self.persist_db.clear_cache()

    def set_raw(self, key: str, value):
        """
        Set data from the DB based on key.

        It will set it both on cache_db and persist_db

        :param key: str:  The key to set, including base_path.
        :param value: The raw, pickled data.
        """
        self.persist_db.set_raw(key, value)
        self.cache_db.set_raw(key, value)
